/**
  ******************************************************************************
  * @file    lcdconf.h
  * @author  MCD Application Team
  * @brief   This file implements the configuration for the GUI library
  ******************************************************************************
  * @attention
  *
  * <h2><center>&copy; Copyright (c) 2019 STMicroelectronics.
  * All rights reserved.</center></h2>
  *
  * This software component is licensed by ST under Ultimate Liberty license
  * SLA0044, the "License"; You may not use this file except in compliance with
  * the License. You may obtain a copy of the License at:
  *                             www.st.com/SLA0044
  *
  ******************************************************************************
  */
  

#ifndef LCDCONF_H
#define LCDCONF_H



#include "GUIDRV_Lin.h"
#include "LCD.h"




typedef struct {
	uint32_t mHactive;
	uint32_t mHfp;
	uint32_t mHbp;
	uint32_t mHsync;
	uint32_t mVactive;
	uint32_t mVfp;
	uint32_t mVbp;
	uint32_t mVsync;
	uint32_t mBitsPerPixel;
	uint32_t mBytesPerPixel;
	uint32_t mLcdRotation;
	uint32_t mTouchRotation;
	uint32_t xMax;
	uint32_t yMax;
} VideoTimingType;

extern const VideoTimingType microtec1280x800_lcd;


#undef  LCD_SWAP_XY
#undef  LCD_MIRROR_Y


#define LCD_SWAP_XY  1 //1
#define LCD_MIRROR_Y 1 //1

#undef  LCD_SWAP_RB
#define LCD_SWAP_RB  1 //1

/* Define the possible screen orientations  */
#define ROTATION_0   0
#define ROTATION_CW  1
#define ROTATION_180 2
#define ROTATION_CCW 3


/* Define the possible color modes */
#define CMS_ARGB8888 1
#define CMS_RGB888   2
#define CMS_RGB565   3
#define CMS_ARGB1555 4
#define CMS_ARGB4444 5
#define CMS_L8       6
#define CMS_AL44     7
#define CMS_AL88     8


#define XSIZE_PHYS 		1280//800
#define YSIZE_PHYS 		800//480

#define SINGLE_BUFFER_SIZE		(XSIZE_PHYS * YSIZE_PHYS)/// (4 / COLOR_DEPTH_BYTES)


/* Define Layer 0 color mode, orientation and sizes */
#define COLOR_MODE_0  CMS_ARGB8888 // CMS_ARGB8888
#define ORIENTATION_0 ROTATION_0   //ROTATION_180
#define XSIZE_0       XSIZE_PHYS 
#define YSIZE_0       YSIZE_PHYS

/* Define Layer 1 color mode, orientation and sizes */
#define COLOR_MODE_1  CMS_ARGB8888
#define ORIENTATION_1 ROTATION_0 
#define XSIZE_1       XSIZE_PHYS
#define YSIZE_1       YSIZE_PHYS


const VideoTimingType * LCD_GetConfiguration(void);

typedef struct
{
  int32_t      address;          
  __IO int32_t pending_buffer;   
  int32_t      buffer_index;     
  int32_t      xSize;            
  int32_t      ySize;            
  int32_t      BytesPerPixel;
  LCD_API_COLOR_CONV   *pColorConvAPI;
} LCD_LayerPropTypedef;

#endif /* LCDCONF_H */

/*************************** End of file ****************************/
